﻿//Josip Medved <jmedved@jmedved.com> http://www.jmedved.com

//2008-03-04: Initial version.


using System;
using System.Collections.Generic;
using System.Drawing;

namespace Medo.Drawing {

    /// <summary>
    /// Secondary buffer for drawing.
    /// </summary>
    public class SecondaryBuffer : IDisposable {

        private Dictionary<Size, Bitmap> _bitmapBuffer = new Dictionary<Size, Bitmap>();
        private Bitmap _currentBitmap;
        private Bitmap _newBitmap;

        private Dictionary<Size, Graphics> _graphicsBuffer = new Dictionary<Size, Graphics>();
        private System.Drawing.Graphics _newGraphics;


        /// <summary>
        /// Creates new instance.
        /// </summary>
        public SecondaryBuffer() { 
        }


        /// <summary>
        /// Begins drawing to new bitmap.
        /// </summary>
        /// <param name="size">Size of bitmap.</param>
        /// <param name="backgroundColor">Background color.</param>
        public void BeginNewGraphics(Size size, Color backgroundColor) {
            if (_newBitmap != null) { throw new System.InvalidOperationException("BeginNewGraphics without EndNewGraphics."); }

            if (_bitmapBuffer.ContainsKey(size)) {
                _newBitmap = _bitmapBuffer[size];
                _newGraphics = _graphicsBuffer[size];
            } else {
                _newBitmap = new Bitmap(size.Width, size.Height);
                _newGraphics = Graphics.FromImage(_newBitmap);
                _bitmapBuffer.Add(size, _newBitmap);
                _graphicsBuffer.Add(size, _newGraphics);
            }

            _newGraphics.Clip = new Region(new Rectangle(0, 0, _newBitmap.Width, _newBitmap.Height));
            _newGraphics.Clear(backgroundColor);
        }

        /// <summary>
        /// Gets Graphics object for new bitmap.
        /// </summary>
        public Graphics NewGraphics {
            get {
                return this._newGraphics;
            }
        }
        
        /// <summary>
        /// Ensures all drawing is finished.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">EndNewGraphics without BeginNewGraphics.</exception>
        public void EndNewGraphics() {
            if (_newBitmap == null) { throw new System.InvalidOperationException("EndNewGraphics without BeginNewGraphics."); }

            _currentBitmap = _newBitmap;
            _newBitmap = null;
            _newGraphics = null;
        }

        /// <summary>
        /// Ensures all drawing is finished and renders it to desired destination.
        /// </summary>
        /// <param name="destinationGraphics">Graphics on which rendering is to take place.</param>
        /// <exception cref="System.InvalidOperationException">EndNewGraphics without BeginNewGraphics.</exception>
        public void EndNewGraphics(Graphics destinationGraphics) {
            EndNewGraphics();
            Render(destinationGraphics);
        }
        

        /// <summary>
        /// Moves last buffered bitmap to destination.
        /// </summary>
        /// <param name="destinationGraphics">Graphics on which rendering is to take place.</param>
        /// <exception cref="System.NullReferenceException">No buffered bitmap.</exception>
        public void Render(Graphics destinationGraphics) {
            if (this._currentBitmap == null) { throw new NullReferenceException("No buffered bitmap."); }
            destinationGraphics.DrawImage(this._currentBitmap, 0, 0);
        }


        /// <summary>
        /// Releases all buffers not in use.
        /// </summary>
        /// <exception cref="System.InvalidOperationException">Cannot dispose before EndNewGraphics is called.</exception>
        public void ReleaseBuffers() {
            if (_newBitmap != null) { throw new System.InvalidOperationException("Cannot dispose before EndNewGraphics is called."); }

            using (Dictionary<Size, Graphics>.Enumerator eGraphics = _graphicsBuffer.GetEnumerator()) {
                while (eGraphics.MoveNext()) {
                    eGraphics.Current.Value.Dispose();
                }
            }
            using (Dictionary<Size, Bitmap>.Enumerator eBitmap = _bitmapBuffer.GetEnumerator()) {
                while (eBitmap.MoveNext()) {
                    if (eBitmap.Current.Value != _currentBitmap) {
                        eBitmap.Current.Value.Dispose();
                    }
                }
            }
            _newBitmap = null; //no need to dispose - it is already disposed
            _newGraphics = null;
        }


        #region IDisposable Members

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        public void Dispose() {
            this.Dispose(true);
            System.GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        /// <param name="disposing">True if managed resources should be disposed; otherwise, false.</param>
        protected virtual void Dispose(bool disposing) {
            if (disposing) {
                using (Dictionary<Size, Graphics>.Enumerator eGraphics = _graphicsBuffer.GetEnumerator()) {
                    while (eGraphics.MoveNext()) {
                        eGraphics.Current.Value.Dispose();
                    }
                }
                using (Dictionary<Size, Bitmap>.Enumerator eBitmap = _bitmapBuffer.GetEnumerator()) {
                    while (eBitmap.MoveNext()) {
                        eBitmap.Current.Value.Dispose();
                    }
                }

                _newBitmap = null; //no need to dispose - it is already disposed
                _newGraphics = null;
                _currentBitmap = null;
                _bitmapBuffer = null;
                _graphicsBuffer = null;
            }
        }

        #endregion
    }
}
